import cPickle as pickle

class NaiveBayesClassifier(dict):
    """
    Namespaces classifiers so that we can refer to them individually

    Simple wrapper around the dict object
    """
    def __init__(self, *args, **kwargs):
        super(NaiveBayesClassifier, self).__init__(*args, **kwargs)

    def _getident(self):
        return self._ident
    ident = property(_getident)

    @classmethod
    def load(cls, file_path):
        return cls(pickle.load(open(file_path, 'rb')))

class FMClassifier(NaiveBayesClassifier):
    """
    Federated Media Classifier
    """
    def __init__(self, *args, **kwargs):
        self._ident = 'fm'
        super(NaiveBayesClassifier, self).__init__(*args, **kwargs)




#!/usr/bin/env python
import cPickle as pickle
import json
import time
from backends import RedisBackend
from bayes import NaiveBayes
from classifiers import FMClassifier
from optparse import OptionParser
parser = OptionParser(conflict_handler='resolve')
parser.add_option('-h', dest='host')
parser.add_option('-p', '--port', dest='port')
options, args = parser.parse_args()

clsfr = FMClassifier(pickle.load(open('/Users/georgecourtsunis/projects/disqus/disqus/analytics/hadoop/thread_views/var/cmap.b', 'rb')))
backend = RedisBackend(host=options.host, port=options.port)
bayes = NaiveBayes(backend=backend)

_start = time.time()
for file_name in args:
    print 'Training file %s' % file_name
    fd = open(file_name, 'r')
    _counter = 0
    for line in fd:
        _counter += 1
        if _counter % 100000 == 0:
            print _counter, (time.time() - _start)
        # if _counter % 1000000 == 0:
        #     break
        # grab args
        vector, count = line.split('\t')
        vector = json.loads(vector)
        count = int(count)
        bayes.train([vector] * count)

print 'Trained in %0.2f seconds' % (time.time() - _start)




# import sys
# import time
from collections import defaultdict as dd

mean = lambda x: sum(x) / len(x)

exact_match = lambda w, c: w == c
starts_match = lambda w, c: w.startswith(c) and len(c) > (len(w) / 2)
word_in_cat = lambda w, c: w in c
cat_in_word = lambda w, c: c in w

stop_words = set([
    'or', 'of', 'and', 'to', '&', 'back', 'new', 'personal',
    'power', 'care', 'back', 'loss', 'day', 'video',
])

class NaiveBayes(object):
    """
    Simple implementation of a naive Bayes classifier.
    """
    def __init__(self, backend=None):
        if backend is None:
            # TODO fillin with a default connection or something
            raise Exception, 'Hey asshole, supply a backend'
        self.b = backend

    def _calc_feature_score(self, classifier, feature, scale=1):
        """
        Given a feature and classifier, determine the score
        """
        r = 0
        # normalize the feature
        feature = feature.lower()
        conds = (
            exact_match,
            starts_match,
            word_in_cat,
            cat_in_word,
        )
        for cond in conds:
            # shortcut exact to hit sets
            if cond == exact_match:
                if feature == classifier['name']:
                    r += 1 * scale
                elif feature in classifier['parts_s'].difference(stop_words):
                    r += 1 * scale * 0.95
                elif feature in classifier['hyponyms_s']:
                    r += 1 * scale * 0.90
                elif feature in classifier['hypernyms_s']:
                    r += 1 * scale * 0.90
                else:
                    continue
            # normal comparisons
            else:
                if cond(feature, classifier['name']):
                    r += 1 * scale
                # check parts
                elif any(cond(feature, c_p) for c_p in classifier['parts'] if c_p not in stop_words):
                    r += 1 * scale
                # TODO each of these checks will slow down classifier
                # # check hyponyms
                # elif any(cond(feature, c_ho) for c_ho in classifier['hyponyms'] if c_ho not in stop_words):
                elif any(cond(feature, c_ho) for c_ho in classifier['hyponyms']):
                    r += 1 * scale * 0.90
                # check hypernyms
                elif any(cond(feature, c_h) for c_h in classifier['hypernyms']):
                    r += 1 * scale * 0.90
                # # check part hyponyms
                # elif any(cond(feature, c) for c in classifier['phyponyms']):
                #     r = 1 * scale * 0.001
                # # check part hypernyms
                # elif any(cond(feature, c) for c in classifier['phypernyms']):
                #     r = 1 * scale * 0.002
                else:
                    continue
        return r + 1e-10 # small K for > 0 multiplication

    def _calc_scores(self, Classifier, Feature, key, **kwargs):
        """
        p(C) = p(f1|c1) * p(f2|c1), etc for all f in F, for all c in C
        """
        # build up a map of all cats/features/scores
        cat_feature_score_map = dd(dict)
        for classifier, meta in Classifier.iteritems():
            meta['name'] = classifier
            # TODO figure out if we want to do something more interesting here
            for i, feature in enumerate(Feature):
                # feature not in cat map
                if feature not in cat_feature_score_map[classifier]:
                    if kwargs.get('linear_weight_vector'):
                        cat_feature_score_map[classifier][feature] = self._calc_feature_score(meta, feature, scale=i+1)
                    else:
                        cat_feature_score_map[classifier][feature] = self._calc_feature_score(meta, feature)
                # feature in cat map
                else:
                    if kwargs.get('linear_weight_vector'):
                        cat_feature_score_map[classifier][feature] += self._calc_feature_score(meta, feature, scale=i+1)
                    else:
                        cat_feature_score_map[classifier][feature] += self._calc_feature_score(meta, feature)
                # train the bayesian based on the situational names
                if cat_feature_score_map[classifier][feature] > 0.50:
                    pass
                    # self.b.update_classification(Classifier.ident, classifier, key)

        # calculate the probabilty distribution
        cat_score_map = dd(float)
        for classifier, features in cat_feature_score_map.iteritems():
            # grab currently probability
            p_cat = self.b.get_classification(Classifier.ident, classifier, key)
            # print classifier, p_cat
            # score = p(c) * p(f1|c1) * p(f2|c1), etc for all f in F, for all c in C
            scores = [f[1] for f in features.iteritems()]
            cat_score_map[classifier] = (p_cat * reduce(lambda x, y: x * y, scores, 1))
        # return a list of probabilities for each category (between 0 and 1)
        return sorted(cat_score_map.iteritems(), key=lambda x:x[1], reverse=True)

    def _calc_classification(self, Classifier, Feature, key, **kwargs):
        scores = self._calc_scores(Classifier, Feature, key, **kwargs)
        return scores[0] # most relevant classification

    def train(self, vectors):
        # vector -> Classifier_id, classifier, key
        for v in vectors:
            self.b.batch_update(*v)
        # persist
        self.b.flush_pipe()

    def classify(self, Classifier, Feature, key, **kwargs):
        classifier, prob = self._calc_classification(Classifier, Feature, key, **kwargs)
        self.b.update_classification(Classifier, classifier, key)
        # print 'Classified as', classifier, prob
        return classifier, prob




#!/usr/bin/env python
import cPickle as pickle
from backends import RedisBackend
from bayes import NaiveBayes
from classifiers import FMClassifier

training_data = (
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'politics', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'business', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'music', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
    ('fm', 'tech news', 'cnn'),
)

cmap = pickle.load(open('../hadoop/thread_views/var/cmap.b',
        'rb'))
clsfr = FMClassifier(cmap)
backend = RedisBackend()
bayes = NaiveBayes(backend=backend)
bayes.train(training_data)
# no we are ready to test the bayes filter

# TODO add support for subfeatures in features

import time
_start = time.time()
# bayes.classify(clsfr, ('aldfksjalskdjfasdflapoliticsadlskfajsldfj',), 'cnn')
# bayes.classify(clsfr, ('politics', 'aldfksjalspoliticskdjfasdflbusinessapmusicadlhomeskfajsldfj', 'music'), 'cnn')
# bayes.classify(clsfr, ('business', 'music', 'love', 'living', 'politics', 'music'), 'cnn', linear_weight_vector=True)
# bayes.classify(clsfr, ('tech', 'computers', 'news'), 'cnn')
line = 'gaming.www.myvidster.com/video/2797926/PornoTubecom_-_Keymon_Phoenix_Mister_Buck_Dee_Truth_Intrigue_and_Jermany_-_Browsin'
bayes.classify(clsfr, line.split('/'), 'myvidster.com', linear_weight_vector=True)
print (time.time() - _start), 'seconds'




from distutils.core import setup

setup(
    name='bayes',
    description='Bayes classification implementation',
    author='George Courtsunis',
    author_email='gjcourt@gmail.com',
    url='http://github.com/gjcourt/bayes',
    version='0.1',
    packages=['bayes']
    )




from redis import Redis

class NaiveBayesBackend(object):

    def __init__(self, host, port):
        if not host:
            host = 'localhost'
        if not port:
            port = 6379
        self.conn = Redis(host=host, port=port)

    def get_classification(self, C, c, k):
        """
        p(c) for the given key
        for a given key and classifier, get the probability
        """
        raise NotImplementedError

    def update_classification(self, C, c, k):
        raise NotImplementedError


class RedisBackend(NaiveBayesBackend):

    def __init__(self, host=None, port=None):
        # call up to init the redis connection
        super(RedisBackend, self).__init__(host, port)
        self._count = 0
        self.pipe = self.conn.pipeline()

    def _get_name(self, C, k):
        return '%s.%s' % (C, k)

    def _get_keys(self, C, c, k):
        return (
            '%s' % c,
            'T',
        )

    def flush_pipe(self):
        self.pipe.execute()
        self.pipe = self.conn.pipeline()

    def get_classification(self, C, c, k):
        name = self._get_name(C, k)
        keys = self._get_keys(C, c, k)
        c_count = float(self.conn.hget(name, keys[0]) or 0)
        C_count = float(self.conn.hget(name, keys[1]) or 0) + 1e-10 # correct for /0 errors
        return c_count / C_count

    def update_classification(self, C, c, k):
        name = self._get_name(C, k)
        keys = self._get_keys(C, c, k)
        self.conn.hincrby(name, keys[0], 1)
        self.conn.hincrby(name, keys[1], 1)

    def batch_update(self, C, c, k):
        name = self._get_name(C, k)
        keys = self._get_keys(C, c, k)
        self.pipe.hincrby(name, keys[0], 1)
        self.pipe.hincrby(name, keys[1], 1)
        self._count += 1
        if self._count % 100000 == 0:
            self.flush_pipe()
            self._count = 0

